/*
 This file is part of the digitalSTROM dsm-api library.

 Copyright (C) 2010 digitalSTROM.org, Schlieren, Switzerland

 digitalSTROM dsm-api library is free software: you can redistribute it and/or modify
 it under the terms of the GNU Lesser General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 digitalSTROM dsm-api library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public License
 along with digitalSTROM dsm-api library. If not, see <http://www.gnu.org/licenses/>.
 */

#include "dsm-api.h"
#include "uthash.h"
#include <digitalSTROM/ds485-client-interface.h>
#include "dsm-macros.h"
#include <stdlib.h>

typedef struct {
    ds485ClientHandle_t client;
    struct callback_item *callback_list;
    BusState_callback_t bus_state_callback;
    void *bus_state_arg;
    BusChange_callback_t bus_change_callback;
    void *bus_change_arg;
} dsm_api_data_t;

static int dsm_api_ds485_message_callback(void *handle, const ds485_container_t *container);
static int dsm_api_ds485_bus_state_callback(void *handle, const bus_state_t state);
static int dsm_api_ds485_bus_change_callback(void *handle, dsid_t *id, int flags);

typedef void (*callback_function_t)();

struct callback_item {
    uint32_t id;
    callback_function_t callback;
    void *arg;
    UT_hash_handle hh;
};

static struct callback_item *callback_list = NULL;

DsmApiHandle_t DsmApiInitialize(void)
{
    dsm_api_data_t *data = (dsm_api_data_t*)malloc(sizeof(dsm_api_data_t));

    if (data == NULL)
    {
        return NULL;
    }

    data->callback_list = NULL;
    data->bus_change_callback = NULL;
    data->bus_state_callback = NULL;

    return data;
}

int DsmApiOpen(DsmApiHandle_t handle, const char *connection, int flags)
{
    dsm_api_data_t *data = (dsm_api_data_t*)handle;

    if (data == NULL)
    {
        return -1;
    }

    data->client = ds485_client_open(connection, flags);
    if (data->client == NULL)
    {
        return -1;
    }

    /* register message callback with lower layer */
    ds485_client_set_receive_callback(data->client, &dsm_api_ds485_message_callback, data);
    ds485_client_set_state_callback(data->client, &dsm_api_ds485_bus_state_callback, data);
    ds485_client_set_bus_change_callback(data->client, &dsm_api_ds485_bus_change_callback, data);

    return 0;
}

int DsmApiClose(DsmApiHandle_t handle)
{
    dsm_api_data_t *data = (dsm_api_data_t*)handle;

    if (data == NULL)
    {
        return -1;
    }

    struct callback_item *item;

    /* clear message callback with lower layer */
    ds485_client_set_receive_callback(data->client, NULL, NULL);

    while (data->callback_list)
    {
        item = data->callback_list;
        HASH_DEL(data->callback_list, item);
        free(item);
    }

    return ds485_client_close(data->client);
}

int DsmApiCleanup(DsmApiHandle_t handle)
{
    dsm_api_data_t *data = (dsm_api_data_t*)handle;

    if (data == NULL)
    {
        return -1;
    }

    free(data);

    return 0;
}

void *DsmApiRegisterCallback(DsmApiHandle_t handle, container_type_t MessageType, uint8_t MessageId, uint8_t ModifierId, void *callback_function, void *arg)
{
    struct callback_item *callback;
    callback_function_t prev_callback = NULL;

    uint32_t id = (MessageType << 16) | (MessageId << 8) | (ModifierId);

    dsm_api_data_t *data = (dsm_api_data_t*)handle;

    if (data == NULL)
    {
        return NULL;
    }

    HASH_FIND_INT(data->callback_list, &id, callback);

    if (callback == NULL)
    {
        callback = (struct callback_item *)malloc(sizeof(struct callback_item));

        callback->id = id;
        callback->callback = (callback_function_t)callback_function;
        callback->arg = arg;
        HASH_ADD_INT(data->callback_list, id, callback);
    }
    else
    {
        prev_callback = callback->callback;
        callback->callback = (callback_function_t)callback_function;
        callback->arg = arg;
        /* no need to add, as it's already part of the hash */
    }

    return prev_callback;
}

void *DsmApiRegisterBusStateCallback(DsmApiHandle_t handle, BusState_callback_t callback_function, void *arg)
{
    dsm_api_data_t *data = (dsm_api_data_t*)handle;

    if (data == NULL)
    {
        return NULL;
    }

    BusState_callback_t prev_cb = data->bus_state_callback;
    data->bus_state_callback = callback_function;
    data->bus_state_arg = arg;
    return prev_cb;
}

void *DsmApiRegisterBusChangeCallback(DsmApiHandle_t handle, BusChange_callback_t callback_function, void *arg)
{
    dsm_api_data_t *data = (dsm_api_data_t*)handle;

    if (data == NULL)
    {
        return NULL;
    }

    BusChange_callback_t prev_cb = data->bus_change_callback;
    data->bus_change_callback = callback_function;
    data->bus_change_arg = arg;
    return prev_cb;
}

int DsmApiGetBusMembers(DsmApiHandle_t handle, dsid_t *device_list, int device_list_size)
{
    dsm_api_data_t *data = (dsm_api_data_t*)handle;

    if (data == NULL)
    {
        return -1;
    }

    return ds485_client_query_devices(data->client, device_list, device_list_size);
}

int DsmApiGetBusState(DsmApiHandle_t handle, bus_state_t *state)
{
    dsm_api_data_t *data = (dsm_api_data_t*)handle;

    if (data == NULL)
    {
        return -1;
    }

    return ds485_client_query_state(data->client, state);
}

static int dsm_api_ds485_bus_state_callback(void *handle, const bus_state_t state)
{
    dsm_api_data_t *data = (dsm_api_data_t*)handle;

    if (data == NULL)
    {
        return -1;
    }

    if (data->bus_state_callback != NULL)
    {
        data->bus_state_callback(data->bus_state_arg, state);
    }
}

static int dsm_api_ds485_bus_change_callback(void *handle, dsid_t *id, int flags)
{
    dsm_api_data_t *data = (dsm_api_data_t*)handle;

    if (data == NULL)
    {
        return -1;
    }

    if (data->bus_change_callback != NULL)
    {
        data->bus_change_callback(data->bus_change_arg, id, flags);
    }
}

int DsmApiGetOwnDSID(DsmApiHandle_t handle, dsid_t *dsid)
{
    dsm_api_data_t *data = (dsm_api_data_t*)handle;

    if (data == NULL)
    {
        return -1;
    }

    return ds485_client_get_dsid(data->client, dsid);
}
