/*
 This file is part of the digitalSTROM dsm-api library.

 Copyright (C) 2010 digitalSTROM.org, Schlieren, Switzerland

 digitalSTROM dsm-api library is free software: you can redistribute it and/or modify
 it under the terms of the GNU Lesser General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 digitalSTROM dsm-api library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public License
 along with digitalSTROM dsm-api library. If not, see <http://www.gnu.org/licenses/>.
 */

#include "dsm-api.h"
#include "uthash.h"
#include <digitalSTROM/ds485-client-interface.h>
#include "dsm-macros.h"
#include <stdlib.h>

typedef struct {
    ds485ClientHandle_t client;
    struct callback_item *callback_list;
} dsm_api_data_t;

static int dsm_api_ds485_message_callback(void *handle, const ds485_container_t *container);

typedef void (*callback_function_t)();

struct callback_item {
    uint32_t id;
    callback_function_t callback;
    void *arg;
    UT_hash_handle hh;
};

static struct callback_item *callback_list = NULL;

DsmApiHandle_t DsmApi1Initialize(void)
{
    dsm_api_data_t *data = (dsm_api_data_t*)malloc(sizeof(dsm_api_data_t));

    if (data == NULL)
    {
        return NULL;
    }

    data->callback_list = NULL;

    return data;
}

int DsmApi1Open(DsmApiHandle_t handle, const char *connection, int flags)
{
    dsm_api_data_t *data = (dsm_api_data_t*)handle;

    if (data == NULL)
    {
        return -1;
    }

    data->client = ds485_client_open(connection, flags);
    if (data->client < 0)
    {
        return -1;
    }

    /* register message callback with lower layer */
    ds485_client_set_receive_callback(data->client, &dsm_api_ds485_message_callback, data);

    return 0;
}

int DsmApi1Close(DsmApiHandle_t handle)
{
    dsm_api_data_t *data = (dsm_api_data_t*)handle;

    if (data == NULL)
    {
        return -1;
    }

    struct callback_item *item;

    /* clear message callback with lower layer */
    ds485_client_set_receive_callback(data->client, NULL, NULL);

    while (data->callback_list)
    {
        item = data->callback_list;
        HASH_DEL(data->callback_list, item);
        free(item);
    }

    return ds485_client_close(data->client);
}

int DsmApi1Cleanup(DsmApiHandle_t handle)
{
    dsm_api_data_t *data = (dsm_api_data_t*)handle;

    if (data == NULL)
    {
        return -1;
    }

    free(data);

    return 0;
}

void *DsmApiRegisterCallback(DsmApiHandle_t handle, container_type_t MessageType, uint8_t MessageId, uint8_t ModifierId, void *callback_function, void *arg)
{
    struct callback_item *callback;
    callback_function_t prev_callback = NULL;

    uint32_t id = (MessageType << 16) | (MessageId << 8) | (ModifierId);

    dsm_api_data_t *data = (dsm_api_data_t*)handle;

    if (data == NULL)
    {
        return NULL;
    }

    HASH_FIND_INT(data->callback_list, &id, callback);

    if (callback == NULL)
    {
        callback = (struct callback_item *)malloc(sizeof(struct callback_item));

        callback->id = id;
        callback->callback = (callback_function_t)callback_function;
        callback->arg = arg;
        HASH_ADD_INT(data->callback_list, id, callback);
    }
    else
    {
        prev_callback = callback->callback;
        callback->callback = (callback_function_t)callback_function;
        callback->arg = arg;
        /* no need to add, as it's already part of the hash */
    }

    return prev_callback;
}
